structure Eval :> EVAL =
struct

  open MinML
  structure S = Stream


  (**** Substitution ****)

  (* id is the identity substitution (does nothing). *)

  val id = fn x => Var(x)

  (* r++(x,e) is the substitution taking x to e and applying r
       to all other variables. *)

  infix 1 ++
  fun (r ++ (x,e)) = fn x' => if x = x' then e else r(x')

  (* csubst(r,e) applies the CLOSED substitution r to the
       expression e. *)

  fun csubst (r,Var(x)) = r(x)
    | csubst (r,Fun(f,x,t,t',e)) =
        Fun(f,x,t,t',csubst(r++(f,Var(f))++(x,Var(x)),e))
    | csubst (r,Apply(e,e')) = Apply(csubst(r,e),csubst(r,e'))
    | csubst (r,If(e,e1,e2)) = If(csubst(r,e),csubst(r,e1),csubst(r,e2))
    | csubst (r,Primop(p,es)) = Primop(p, map (fn e => csubst(r,e)) es)
    | csubst (r,e) = e



  (**** Evaluation ****)

  (* value(e) = true iff e is a value *)

  fun value (Int _) = true
    | value (Bool _) = true
    | value (Fun _) = true
    | value _ = false

  exception Stuck and Done

  (* step(e) = e' for some e' such that e -> e'
             = raise Stuck if no such e' exists. *)

  fun step (If(Bool true,e,_)) = e
    | step (If(Bool false,_,e)) = e
    | step (If(e,e1,e2)) = If(step(e),e1,e2)

    | step (Primop(p,es)) =
          (Primop(p,stepList(es)) handle Done =>
               (case evalPrimop(p,es)
                  of NONE => raise Stuck
                   | SOME e => e))

    | step (Apply(e,e')) =
          if not(value(e)) then Apply(step(e),e')
          else if not(value(e')) then Apply(e,step(e'))
          else (case e
                  of (Fun(f,x,_,_,e'')) => csubst(id++(f,e)++(x,e'),e'')
                   | _ => raise Stuck)

    | step _ = raise Stuck

  (* stepList(l) applies step to the first element of l which is
       not a value, or raises Done if every element of l is a value. *)

  and stepList [] = raise Done
    | stepList (e::es) = if not(value(e)) then step(e)::es
                           else e::stepList(es)

  fun stepOpt e = SOME (step e) handle Stuck => NONE

  fun multiStep e = multiStep (step e) handle Stuck => e

  (* Return a stream containing all the steps of an evaluation. *)

  fun stepStream e =
      let fun steps e =
              (case stepOpt e
                 of NONE => S.empty
                  | SOME e' => S.delay (fn () => S.Cons(e', steps e')))
       in S.delay (fn () => S.Cons(e, steps e)) end

end
